#!/usr/bin/env python3

#  This file is part of Leela Chess Zero.
#  Copyright (C) 2020 The LCZero Authors
#
#  Leela Chess is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Leela Chess is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Leela Chess.  If not, see <http://www.gnu.org/licenses/>.
#
#  Additional permission under GNU GPL version 3 section 7
#
#  If you modify this Program, or any covered work, by linking or
#  combining it with NVIDIA Corporation's libraries from the NVIDIA CUDA
#  Toolkit and the NVIDIA CUDA Deep Neural Network library (or a
#  modified version of those libraries), containing parts covered by the
#  terms of the respective license agreement, the licensors of this
#  Program grant you additional permission to convey the resulting work.

import sys
from pybind.writer import Writer
from pybind import Module, Class
from pybind.parameters import (StringParameter, ClassParameter,
                               NumericParameter, ArgvObjects, IntegralArgv,
                               ListOfStringsParameter)
from pybind.retval import (StringViewRetVal, StringRetVal, ListOfStringsRetVal,
                           NumericRetVal, ObjCopyRetval, ObjOwnerRetval,
                           ObjTupleRetVal, IntegralTupleRetVal)
from pybind.exceptions import CppException

# Module
mod = Module('backends')
mod.AddInclude('python/weights.h')
mod.AddInitialization('lczero::InitializeMagicBitboards();')
ex = mod.AddException(
    CppException('LczeroException', cpp_name='lczero::Exception'))

# Weights class
weights = mod.AddClass(Class('Weights', cpp_name='lczero::python::Weights'))
weights.constructor.AddParameter(
    StringParameter('filename', optional=True, can_be_none=True)).AddEx(ex)
weights.AddMethod('filename').AddRetVal(StringViewRetVal())
weights.AddMethod('license').AddRetVal(StringViewRetVal())
weights.AddMethod('min_version').AddRetVal(StringRetVal())
weights.AddMethod('input_format').AddRetVal(NumericRetVal('i'))
weights.AddMethod('policy_format').AddRetVal(NumericRetVal('i'))
weights.AddMethod('value_format').AddRetVal(NumericRetVal('i'))
weights.AddMethod('moves_left_format').AddRetVal(NumericRetVal('i'))
weights.AddMethod('blocks').AddRetVal(NumericRetVal('i'))
weights.AddMethod('filters').AddRetVal(NumericRetVal('i'))

# Input class
input = mod.AddClass(Class('Input', cpp_name='lczero::python::Input'))
input.AddMethod('set_mask').AddParameter(NumericParameter('plane'),
                                         NumericParameter(
                                             'mask', type='u64')).AddEx(ex)
input.AddMethod('set_val').AddParameter(NumericParameter('plane'),
                                        NumericParameter('value',
                                                         type='f32')).AddEx(ex)
input.AddMethod('mask').AddParameter(NumericParameter('plane')).AddRetVal(
    NumericRetVal('u64')).AddEx(ex)
input.AddMethod('val').AddParameter(NumericParameter('plane')).AddRetVal(
    NumericRetVal('f32')).AddEx(ex)
input.AddMethod('clone').AddRetVal(ObjOwnerRetval(input))

# Output class
output = mod.AddClass(
    Class('Output',
          cpp_name='lczero::python::Output',
          disable_constructor=True))
output.AddMethod('q').AddRetVal(NumericRetVal('f32'))
output.AddMethod('d').AddRetVal(NumericRetVal('f32'))
output.AddMethod('m').AddRetVal(NumericRetVal('f32'))
output.AddMethod('p_raw').AddParameter(IntegralArgv('samples', 'i')).AddRetVal(
    IntegralTupleRetVal('f32')).AddEx(ex)
output.AddMethod('p_softmax').AddParameter(IntegralArgv(
    'samples', 'i')).AddRetVal(IntegralTupleRetVal('f32')).AddEx(ex)

# Backend capabilities class
backend_caps = mod.AddClass(
    Class('BackendCapabilities',
          cpp_name='lczero::python::BackendCapabilities',
          disable_constructor=True))
backend_caps.AddMethod('input_format').AddRetVal(NumericRetVal('i'))
backend_caps.AddMethod('moves_left_format').AddRetVal(NumericRetVal('i'))

# Backend class
backend = mod.AddClass(Class('Backend', cpp_name='lczero::python::Backend'))
backend.AddStaticMethod('available_backends').AddRetVal(ListOfStringsRetVal())
backend.constructor.AddParameter(
    ClassParameter(weights, 'weights', optional=True),
    StringParameter('backend', optional=True, can_be_none=True),
    StringParameter('options', optional=True, can_be_none=True)).AddEx(ex)
backend.AddMethod('evaluate').AddParameter(ArgvObjects(
    'inputs', input)).AddRetVal(ObjTupleRetVal(output)).AddEx(ex)
backend.AddMethod('capabilities').AddRetVal(ObjCopyRetval(backend_caps))

# PositionHistory class
game_state = mod.AddClass(
    Class('GameState', cpp_name='lczero::python::GameState'))
game_state.constructor.AddParameter(
    StringParameter('fen', optional=True, can_be_none=True),
    ListOfStringsParameter('moves', optional=True),
).AddEx(ex)
game_state.AddMethod('as_input').AddParameter(
    ClassParameter(backend, 'backend',
                   optional=False)).AddRetVal(ObjOwnerRetval(input)).AddEx(ex)
game_state.AddMethod('moves').AddRetVal(ListOfStringsRetVal())
game_state.AddMethod('policy_indices').AddRetVal(IntegralTupleRetVal('i'))
game_state.AddMethod('as_string').AddRetVal(StringRetVal())

with open(sys.argv[1], 'wt') as f:
    writer = Writer(f)
    mod.Generate(writer)


#!/usr/bin/env python3

import argparse
import os


VERSION_FILE = os.path.join(os.path.dirname(os.path.realpath(__file__)), "../src/version.inc")
VERSION_CONTENT = """
#define LC0_VERSION_MAJOR {}
#define LC0_VERSION_MINOR {}
#define LC0_VERSION_PATCH {}
#define LC0_VERSION_POSTFIX "{}"
"""
VERSION_CONTENT = VERSION_CONTENT.strip()


def get_version():
    with open(VERSION_FILE, 'r') as f:
        major = int(f.readline().split()[2])
        minor = int(f.readline().split()[2])
        patch = int(f.readline().split()[2])
        postfix = f.readline().split()[2]

    postfix = postfix.replace('"', '')
    return major, minor, patch, postfix


def set_version(major, minor, patch, postfix=""):
    version_inc = VERSION_CONTENT.format(major, minor, patch, postfix)

    with open(VERSION_FILE, 'w') as f:
        f.write(version_inc)


def update(major, minor, patch, postfix=""):
    set_version(major, minor, patch, postfix)


def main(argv):
    major, minor, patch, postfix = get_version()

    if argv.major:
        major += 1
        minor = 0
        patch = 0
        postfix = ""
        update(major, minor, patch)
    if argv.minor:
        minor += 1
        patch = 0
        postfix = ""
        update(major, minor, patch)
    if argv.patch:
        patch += 1
        postfix = ""
        update(major, minor, patch)
    if argv.postfix and len(argv.postfix) > 0:
        postfix = argv.postfix
        update(major, minor, patch, postfix)

    if len(postfix) == 0:
        print('v{}.{}.{}'.format(major, minor, patch))
    else:
        print('v{}.{}.{}-{}'.format(major, minor, patch, postfix))


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description=\
            'Set or read current version.')
    argparser.add_argument('--major', action='store_true',
            help='bumps major version')
    argparser.add_argument('--minor', action='store_true',
            help='bumps minor version')
    argparser.add_argument('--patch', action='store_true',
            help='bumps patch')
    argparser.add_argument('--postfix', type=str,
            help='set postfix')
    argv = argparser.parse_args()
    main(argv)




#!/usr/bin/env python3

#  This file is part of Leela Chess Zero.
#  Copyright (C) 2020 The LCZero Authors
#
#  Leela Chess is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Leela Chess is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Leela Chess.  If not, see <http://www.gnu.org/licenses/>.
#
#  Additional permission under GNU GPL version 3 section 7
#
#  If you modify this Program, or any covered work, by linking or
#  combining it with NVIDIA Corporation's libraries from the NVIDIA CUDA
#  Toolkit and the NVIDIA CUDA Deep Neural Network library (or a
#  modified version of those libraries), containing parts covered by the
#  terms of the respective license agreement, the licensors of this
#  Program grant you additional permission to convey the resulting work.

import argparse
import os
import re
import sys

VARINT_TYPES = {
    'int32': 'std::int32_t',
    'int64': 'std::int64_t',
    'uint32': 'std::uint32_t',
    'uint64': 'std::uint64_t',
    'sint32': 'std::uint32_t',
    'sint64': 'std::uint64_t',
    'bool': 'bool',
}

FIXED64_TYPES = {
    'fixed64': 'std::uint64_t',
    'sfixed64': 'std::int64_t',
    'double': 'double',
}
FIXED32_TYPES = {
    'fixed32': 'std::uint32_t',
    'sfixed32': 'std::int32_t',
    'float': 'float',
}
BYTES_TYPES = {
    'string': 'std::string_view',
    'bytes': 'std::string_view',
}
ZIGZAG_TYPES = set(['sint32', 'sint64'])
FLOAT_TYPES = set(['float', 'double'])

TYPES = {**VARINT_TYPES, **FIXED32_TYPES, **FIXED64_TYPES, **BYTES_TYPES}

RESERVED_WORDS = [
    'syntax',
    'package',
    'message',
    'optional',
    'repeated',
    'enum',
] + list(TYPES.keys())

GRAMMAR = ([(r'%s\b' % x, x)
            for x in RESERVED_WORDS] + [('\\' + x, x) for x in '=;{}.'] + [
                (r'/\*.*?\*/', None),  # /* Comment */
                (r'//.*?$', None),  # // Comment
                (r'\s+', None),  # Whitespace
                (r'$', 'EOF'),
                (r'"((?:[^"\\]|\\.)*)"', 'string'),
                (r'\d+', 'number'),
                (r'\w+', 'identifier'),
            ])


class Lexer:
    def __init__(self, text):
        self.text = text
        self.grammar = [(re.compile(x, re.S + re.M), y) for x, y in GRAMMAR]
        self.cur_token = None
        self.cur_offset = 0

    def Pick(self):
        '''Picks the last token in queue. Doesn't advance the queue.'''
        if self.cur_token is None:
            self.cur_token = self.NextToken()
        return self.cur_token

    def Consume(self, expected_token, value=None, group=0):
        '''Gets the token from the queue and advances the queue.

        If @expected_token if of wrong type, or @value is not equal to regexes
        @group, throws an error.
        '''
        token, match = self.Pick()
        if expected_token != token:
            self.Error('Expected token type [%s]' % expected_token)
        if value is not None and value != match.group(group):
            self.Error('Expected value [%s]' % value)
        self.cur_offset = match.span()[1]
        self.cur_token = None
        return match

    def NextToken(self):
        '''Reads the stream and returns the next token.

        (which is not whitespace or comment)
        '''
        while True:
            token, match = self.NextTokenOrWhitespace()
            if token is None:
                self.cur_offset = match.span()[1]
            else:
                return token, match

    def NextTokenOrWhitespace(self):
        '''Reads the stream and returns the next token (possibly whitespace).'''
        for r, token in self.grammar:
            m = r.match(self.text, self.cur_offset)
            if m:
                return (token, m)
        self.Error('Unexpected token')

    def Error(self, text):
        '''Throws an error with context in the file read.'''
        line_start = self.text.rfind('\n', 0, self.cur_offset) + 1
        line_end = self.text.find('\n', line_start)
        sys.stderr.write('%s:\n' % text)
        sys.stderr.write(self.text[line_start:line_end] + '\n')
        sys.stderr.write(' ' * (self.cur_offset - line_start) + '^^^\n')
        raise ValueError("Parse error: %s at offset %d." %
                         (text, self.cur_offset))


def ReadIdentifierPath(lexer):
    '''Reads qualified identifier a.b.d into ['a', 'b', 'd'] list'''
    path = []
    while True:
        path.append(lexer.Consume('identifier').group(0))
        if lexer.Pick()[0] != '.':
            return path
        lexer.Consume('.')


def LookupType(name, stack):
    '''Looks up the (possibly qualified) from the innermost scope first.'''
    for y in stack:
        for x in y:
            if x.GetName() == name[0]:
                if len(name) == 1:
                    return x.GetType()
                else:
                    return LookupType(name[1:], [x.GetTypes()])
    raise ValueError("Cannot find type: %s." % '.'.join(name))


# All *Parser classes have the following semantics:
# * They are called with lexer as input to parse grammar from .proto file.
# * The Generate() function writes relevant portion of .pb.h file.


class ProtoTypeParser:
    def __init__(self, lexer, object_stack):
        token, match = lexer.Pick()
        if token in TYPES:
            self.typetype = 'basic'
            self.name = token
            lexer.Consume(token)
        elif token == 'identifier':
            self.name = ReadIdentifierPath(lexer)
            self.typetype = LookupType(self.name, object_stack)
        else:
            lexer.Error('Type expected')

    def IsZigzag(self):
        if self.typetype == 'basic':
            return self.name in ZIGZAG_TYPES
        return False

    def GetCppType(self):
        if self.typetype == 'basic':
            return TYPES[self.name]
        else:
            return '::'.join(self.name)

    def GetVariableCppType(self):
        if self.IsBytesType():
            return 'std::string'
        else:
            return self.GetCppType()

    def IsVarintType(self):
        return self.typetype == 'enum' or (self.typetype == 'basic'
                                           and self.name in VARINT_TYPES)

    def IsFixedType(self):
        return self.typetype == 'basic' and (self.name in FIXED64_TYPES
                                             or self.name in FIXED32_TYPES)

    def IsBytesType(self):
        return self.typetype == 'basic' and self.name in BYTES_TYPES

    def IsFloatType(self):
        return self.typetype == 'basic' and self.name in FLOAT_TYPES

    def GetWireType(self):
        if self.typetype == 'basic':
            if self.name in VARINT_TYPES:
                return 0
            if self.name in FIXED64_TYPES:
                return 1
            if self.name in BYTES_TYPES:
                return 2
            if self.name in FIXED32_TYPES:
                return 5
            raise ValueError('Unknown type %s' % self.name)
        elif self.typetype == 'enum':
            return 0
        elif self.typetype == 'message':
            return 2
        else:
            raise ValueError('Unknown typetype %s' % self.typetype)

    def IsMessage(self):
        return self.typetype == 'message'

    def IsIntegralType(self):
        if self.typetype == 'basic':
            if self.name == 'double':
                return False
            if self.name == 'float':
                return False
            if self.name in BYTES_TYPES:
                return False
            if self.name in TYPES:
                return True
            raise ValueError('Unknown type %s' % self.name)
        elif self.typetype == 'enum':
            return True
        elif self.typetype == 'message':
            return False
        else:
            raise ValueError('Unknown typetype %s' % self.typetype)


class ProtoFieldParser:
    def __init__(self, lexer, object_stack):
        token, match = lexer.Pick()
        if token not in ['repeated', 'optional', 'required']:
            lexer.Error('repeated, optional or required expected')
        self.category = token
        lexer.Consume(token)
        self.type = ProtoTypeParser(lexer, object_stack)
        self.name = lexer.Consume('identifier')
        lexer.Consume('=')
        self.number = int(lexer.Consume('number').group(0))
        lexer.Consume(';')

    def IsType(self):
        return False

    def GetParser(self):
        name = self.name.group(0)
        if self.type.IsMessage():
            if self.category == 'repeated':
                return 'add_%s()->MergeFromString(val)' % name
            else:
                return 'mutable_%s()->MergeFromString(val)' % name

        cpp_type = self.type.GetCppType()
        val = 'NOT IMPLEMENTED!'
        if self.type.IsVarintType():
            val_val = 'UnZigZag(val)' if self.type.IsZigzag() else 'val'
            val = 'static_cast<%s>(%s)' % (cpp_type, val_val)
        elif self.type.IsFixedType():
            if self.type.IsFloatType():
                val = 'bit_cast<%s>(val)' % cpp_type
            else:
                val = 'static_cast<%s>(val)' % cpp_type
        elif self.type.IsBytesType():
            val = 'val'

        if self.category == 'repeated':
            return '%s_.emplace_back(%s)' % (name, val)
        else:
            return 'set_%s(%s)' % (name, val)

    def GenerateCaseClause(self, w):
        w.Write('case %d: %s; break;' % (self.number, self.GetParser()))

    def GenerateClear(self, w):
        name = self.name.group(0)
        if self.category == 'repeated':
            w.Write('%s_.clear();' % name)
        else:
            w.Write('has_%s_ = false;' % name)
            w.Write('%s_ = {};' % name)

    def GenerateOutput(self, w):
        fname = {
            0: 'AppendVarInt',
            1: 'AppendInt64',
            2: 'AppendString',
            5: 'AppendInt32'
        }
        wire_id = self.type.GetWireType()
        if self.category == 'repeated':
            prefix = 'for (const auto& x : %s)' % (self.name.group(0) + '_')
            name = 'x'
        else:
            name = self.name.group(0) + '_'
            prefix = 'if (has_%s)' % (name)
        if self.type.IsMessage():
            name += '.OutputAsString()'
        w.Write('%s %s(%d, %s, &out);' %
                (prefix, fname[wire_id], self.number, name))

    def GenerateFunctions(self, w):
        name = self.name.group(0)
        cpp_type = self.type.GetCppType()
        var_cpp_type = self.type.GetVariableCppType()
        if self.category == 'repeated':
            if self.type.IsMessage():
                w.Write("%s* add_%s() { return &%s_.emplace_back(); }" %
                        (cpp_type, name, name))
            else:
                w.Write("void add_%s(%s val) { %s_.emplace_back(val); }" %
                        (name, cpp_type, name))
            w.Write("const std::vector<%s>& %s() const { return %s_; }" %
                    (var_cpp_type, name, name))
            if self.type.IsMessage():
                w.Write("const %s& %s(size_t idx) const { return %s_[idx]; }" %
                        (cpp_type, name, name))
            else:
                w.Write("%s %s(size_t idx) const { return %s_[idx]; }" %
                        (cpp_type, name, name))
            w.Write("size_t %s_size() const { return %s_.size(); }" %
                    (name, name))
        else:
            w.Write("bool has_%s() const { return has_%s_; }" % (name, name))
            if self.type.IsMessage():
                w.Write("const %s& %s() const { return %s_; }" %
                        (cpp_type, name, name))
                w.Write("%s* mutable_%s() {" % (cpp_type, name))
                w.Indent()
                w.Write('has_%s_ = true;' % (name))
                w.Write('return &%s_;' % name)
                w.Unindent()
                w.Write("}")
            else:
                w.Write("%s %s() const { return %s_; }" %
                        (cpp_type, name, name))
                w.Write("void set_%s(%s val) {" % (name, cpp_type))
                w.Indent()
                w.Write("has_%s_ = true;" % name)
                w.Write("%s_ = val;" % name)
                w.Unindent()
                w.Write("}")

    def GenerateVariable(self, w):
        name = self.name.group(0)
        cpp_type = self.type.GetVariableCppType()
        if self.category == 'repeated':
            w.Write("std::vector<%s> %s_;" % (cpp_type, name))
        else:
            w.Write("bool has_%s_{};" % (name))
            w.Write("%s %s_{};" % (cpp_type, name))
        return


class ProtoEnumParser:
    def __init__(self, lexer):
        lexer.Consume('enum')
        self.name = lexer.Consume('identifier').group(0)
        self.values = []
        lexer.Consume('{')
        while True:
            token, match = lexer.Pick()
            if token == '}':
                break
            key = lexer.Consume('identifier').group(0)
            lexer.Consume('=')
            value = int(lexer.Consume('number').group(0))
            lexer.Consume(';')
            self.values.append((key, value))
        lexer.Consume('}')

    def GetName(self):
        return self.name

    def GetType(self):
        return 'enum'

    def IsType(self):
        return True

    def Generate(self, w):
        # Protobuf enum is mapped directly to C++ enum.
        w.Write('enum %s {' % self.name)
        w.Indent()
        for key, value in self.values:
            w.Write('%s = %d,' % (key, value))
        w.Unindent()
        w.Write('};')


class ProtoMessageParser:
    def __init__(self, lexer, type_stack):
        self.types = []
        self.fields = []
        lexer.Consume('message')
        self.name = lexer.Consume('identifier').group(0)
        lexer.Consume('{')
        while True:
            token, match = lexer.Pick()
            if token == '}':
                break
            elif token == 'message':
                self.types.append(
                    ProtoMessageParser(lexer, [self.types, *type_stack]))
            elif token == 'enum':
                self.types.append(ProtoEnumParser(lexer))
            elif token in ['repeated', 'optional', 'required']:
                self.fields.append(
                    ProtoFieldParser(lexer, [self.types, *type_stack]))
            else:
                lexer.Error('Expected field or type')
        lexer.Consume('}')

    def GetName(self):
        return self.name

    def GetType(self):
        return 'message'

    def IsType(self):
        return True

    def GetTypes(self):
        return self.types

    def GetFieldsGruppedByWireType(self):
        type_to_fields = {}
        for x in self.fields:
            type_to_fields.setdefault(x.type.GetWireType(), []).append(x)
        return type_to_fields

    def WriteFieldParser(self, w, wire_id, fields):
        fname = {0: 'SetVarInt', 1: 'SetInt64', 2: 'SetString', 5: 'SetInt32'}
        tname = {
            0: 'std::uint64_t',
            1: 'std::uint64_t',
            2: 'std::string_view',
            5: 'std::uint32_t'
        }
        w.Write('void %s(int field_id, %s val) override {' %
                (fname[wire_id], tname[wire_id]))
        w.Indent()
        w.Write('switch (field_id) {')
        w.Indent()
        for field in fields:
            field.GenerateCaseClause(w)
        w.Unindent()
        w.Write('}')
        w.Unindent()
        w.Write('}')

    def Generate(self, w):
        # Protobuf message is a C++ class.
        w.Write('class %s : public lczero::ProtoMessage {' % self.name)
        w.Write(' public:')
        w.Indent()
        # Writing submessages and enums.
        for x in self.types:
            x.Generate(w)
        for x in self.fields:
            w.Write('')
            x.GenerateFunctions(w)
        w.Write('')
        w.Write('std::string OutputAsString() const override {')
        w.Indent()
        w.Write('std::string out;')
        for x in sorted(self.fields, key=lambda x: x.number):
            x.GenerateOutput(w)
        w.Write('return out;')
        w.Unindent()
        w.Write('}')
        w.Write('')
        w.Write('void Clear() override {')
        w.Indent()
        for x in self.fields:
            x.GenerateClear(w)
        w.Unindent()
        w.Write('}')
        w.Unindent()
        w.Write('')
        w.Write(' private:')
        w.Indent()
        for k, v in self.GetFieldsGruppedByWireType().items():
            self.WriteFieldParser(w, k, v)
        w.Write('')
        for x in self.fields:
            x.GenerateVariable(w)
        w.Unindent()
        w.Write('};')


class ProtoFileParser:
    '''Root grammar of .proto file'''
    def __init__(self, lexer):
        self.package = None
        self.objects = []
        while True:
            token, match = lexer.Pick()
            if token == 'EOF':
                return
            elif token == 'syntax':
                self.ParseSyntax(lexer)
            elif token == 'package':
                self.ParsePackage(lexer)
            elif token == 'message':
                self.ParseMessage(lexer)
            else:
                lexer.Error('Expected message or something similar')

    def ParseSyntax(self, lexer):
        lexer.Consume('syntax')
        lexer.Consume('=')
        lexer.Consume('string', 'proto2', 1)
        lexer.Consume(';')

    def ParsePackage(self, lexer):
        lexer.Consume('package')
        if self.package is not None:
            lexer.Error('Package was already defined')
        self.package = ReadIdentifierPath(lexer)
        lexer.Consume(';')

    def ParseMessage(self, lexer):
        self.objects.append(ProtoMessageParser(lexer, [self.objects]))

    def Generate(self, w):
        w.Write('// This file is AUTOGENERATED, do not edit.')
        w.Write('#pragma once')
        w.Write('#include "utils/protomessage.h"')
        for x in self.package:
            w.Write('namespace %s {' % x)
        w.Indent()
        for object in self.objects:
            object.Generate(w)
        w.Unindent()
        for x in reversed(self.package):
            w.Write('}  // namespace %s' % x)


class Writer:
    '''A helper class for writing file line by line with indent.'''
    def __init__(self, fo):
        self.fo = fo
        self.indent = 0

    def Indent(self):
        self.indent += 2

    def Unindent(self):
        self.indent -= 2

    def Write(self, text):
        if text:
            self.fo.write(' ' * self.indent + text + '\n')
        else:
            self.fo.write('\n')


if __name__ == "__main__":
    # Have the same flags as protoc has.
    parser = argparse.ArgumentParser(description="Compile protobuf files.")
    parser.add_argument('input', type=str)
    parser.add_argument('--proto_path', type=str)
    parser.add_argument('--cpp_out', type=str)
    args = parser.parse_args()

    rel_path = os.path.relpath(args.input, args.proto_path)
    dest_name = os.path.splitext(rel_path)[0] + '.pb.h'
    dest_path = os.path.join(args.cpp_out, dest_name)
    dest_dir = os.path.dirname(dest_path)
    os.makedirs(dest_dir, exist_ok=True)

    with open(args.input, 'r') as input, open(dest_path, 'w') as output:
        proto_file = ProtoFileParser(Lexer(input.read()))
        writer = Writer(output)
        proto_file.Generate(writer)



from typing import List

import antlr4

import json

from chess_engines.parser.gen import pgnLexer as _lexer, pgnParser as _parser


def moves_to_string(moves: List[str], color: str = 'w', turn: int = 1) -> str:
    """
    Converts the given moves to a string. Example:

    ['e4', 'e5', 'Nf3', 'Nc6'] -> '1. e4 e5 2. Nf3 Nc6'

    :param moves:
        A list containing the moves to convert

    :param color:
        The color to start with (either 'w' or 'b')

    :param turn:
        The turn to start with
    """
    i = 0
    tmp = []
    _moves = moves[:]
    if color == 'b':
        tmp.append(f'{i + turn}..{_moves[0]} ')
        _moves.pop(0)
        i += 1

    for (j, move) in enumerate(_moves):
        if j % 2 == 0:
            tmp.append(f'{i + turn}. {move} ')
            i += 1
        else:
            tmp.append(f'{move} ')

    return ''.join(tmp)[:-1]


class PgnParser:
    """
    This class is an ANTLR4 parser for PGNs. It is possible to read
    and extract information from PGN chess games.
    """

    def __init__(self, pgn: str = None):
        """
        Creates a PGN parser and parses given pgn.
        By default, no pgn is set.
        """
        self.tags = {}
        self.moves = []
        if pgn is not None:
            self.parse_pgn(pgn)

    def parse_pgn(self, pgn: str):
        """
        Parses the given pgn (if given). The result is saved under the attributes
        'tags' (Dict[str, str]) and 'moves' (List[Dict[str, str]).
        The tags are a dictionary that maps each tag to its corresponding
        value, e.g.:

        {
            'White': 'Bobby Fischer',
            'Opening': 'Sicilian Defence'
        }

        The moves are a list of dictionaries with the following properties:

        [
            {
                'move': 'e4',
                'glyph': None,
                'comment: None,
            },
            {
                'move': 'a6',
                'glyph': '?!',
                'comment': 'This is not a good move'
            }
        ]

        """
        pgn_ = antlr4.InputStream(pgn)
        lexer = _lexer.pgnLexer(pgn_)
        stream = antlr4.CommonTokenStream(lexer)
        parser = _parser.pgnParser(stream)
        parser.game()
        self.tags = parser.tags
        self.moves = parser.moves

    def json(self, pgn: str = None) -> str:
        """
        Parses the given pgn (if given) and returns the JSON representation
        of this object as a string.

        :return:
            The JSON representation of this object as a string.
        """
        if pgn is not None:
            self.parse_pgn(pgn)
        return json.dumps({
            'tags': self.tags,
            'moves': self.moves
        }, indent=4)



#  This file is part of Leela Chess Zero.
#  Copyright (C) 2020 The LCZero Authors
#
#  Leela Chess is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Leela Chess is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Leela Chess.  If not, see <http://www.gnu.org/licenses/>.
#
#  Additional permission under GNU GPL version 3 section 7
#
#  If you modify this Program, or any covered work, by linking or
#  combining it with NVIDIA Corporation's libraries from the NVIDIA CUDA
#  Toolkit and the NVIDIA CUDA Deep Neural Network library (or a
#  modified version of those libraries), containing parts covered by the
#  terms of the respective license agreement, the licensors of this
#  Program grant you additional permission to convey the resulting work.


class Writer:
    '''A helper class for writing file line by line with indent.'''
    def __init__(self, fo):
        self.fo = fo
        self.indent = 0

    def Indent(self):
        self.indent += 2

    def Unindent(self):
        self.indent -= 2

    def Write(self, text):
        for line in text.split('\n'):
            if line:
                self.fo.write(' ' * self.indent + line + '\n')
            else:
                self.fo.write('\n')

    def Open(self, text='{'):
        self.Write(text)
        self.Indent()

    def Close(self, text='}'):
        self.Unindent()
        self.Write(text)



import re
import subprocess
import threading
import time
from typing import Union, List, Dict, Any, Callable

import chess

from chess_engines.parser import parser


class UciProtocol:
    """
    This class represents a UCI protocol. Via this class, it is possible
    to communicate with an engine that supports this protocol. The
    constructor requires an executable in order to establish a communication
    with the engine. This communication can be used with the 'cmd' method.
    """

    def __init__(self, exe: str = None):
        """
        Establishes a communication with a chess engine that supports the UCI
        protocol. This connection is available under the attribute 'process'.

        :param exe:
            The path to the engine's executable
        """
        self.process: Union[subprocess.Popen, None] = None
        if exe is not None:
            self.process = subprocess.Popen(exe, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

    def cmd(self, command: str, delay: float = 0, callback: Callable[[str], None] = None) -> str:
        """
        Sends a command to the engine. After that, the output stream is
        read until the engine sends the command 'readyok'. The total output
        is then returned as a string.

        :param command:
            The command to send to the engine

        :param delay:
            The seconds to wait after the command has been sent

        :param callback:
            A function that is called each time the engine prints
            something to the output stream. The only argument of
            the function is a string (the line that was printed).

        :return:
            The engine's output as a string
        """
        self.process.stdin.write(f'{command}\n'.encode())
        self.process.stdin.flush()
        time.sleep(delay)

        tmp = []
        if command != 'quit':
            self.process.stdin.write('isready\n'.encode())
            self.process.stdin.flush()
            while True:
                text = self.process.stdout.readline().strip()
                if callback is not None:
                    callback(str(text))
                if text == 'readyok'.encode():
                    break
                elif text != ''.encode():
                    tmp.append(str(text))
        else:
            tmp.append('readyok')

        return '\n'.join(tmp)


class Engine(UciProtocol):
    """
    This class represents a chess engine. Note that this
    is not an implementation of a chess engine, but an abstract
    class which should be extended by implementations.
    """

    def __init__(self, exe: str = None, fen: str = None, props: Dict[str, Any] = None):
        """
        Initializes this engine with the given FEN position. If no
        position is specified, the position is set to the starting
        chess position.

        :param exe:
            The path to the engine's executable

        :param fen:
            The FEN representation of the position

        :param props:
            The properties of this engine
        """
        super().__init__(exe)

        self.fen: str = 'rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1'
        if fen is not None:
            self.fen = fen

        self.props: Dict[str, Any] = props
        if props is None:
            self.props = {}

    def __del__(self):
        """
        Kills the current process and closes the connection to
        the engine permanently.
        """
        self.cmd('quit')
        self.process.terminate()

    def __enter__(self):
        """
        Enters the 'with' statement.
        :return: self
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exits the 'with' statement. Delegates to '__del__()'.
        """
        self.__del__()

    def ascii(self) -> str:
        """
        Returns an ASCII representation of the current FEN position.

        :return:
            The ASCII representation of the current FEN position
        """
        return chess.Board(self.fen).__str__()

    def play_moves_san(self, moves: Union[List[str], str]):
        """
        Plays a series of moves in SAN notation on the board. No additional actions are taken
        by the engine: This method is meant to update the position only.

        :param moves:
            Either a list of SAN moves (e.g. ['e4', 'e5', 'Nf3', 'Nc6']) or a
            string that represents the moves (e.g. '1. e4 e5 2. Nf3 Nc6'),
            though the former is faster.
        """
        if isinstance(moves, str):
            # parse moves if necessary
            tmp = parser.PgnParser(moves).moves
            moves = []
            for turn in tmp:
                for move in turn:
                    moves.append(move['move'])

        board = chess.Board(self.fen)
        for move in moves:
            if move is not None:
                board.push_san(move)
        self.fen = board.fen()

    def play_moves_uci(self, moves: List[str]):
        """
        Plays a series of moves in UCI notation on the board. No additional actions are taken
        by the engine: This method is meant to update the position only.

        :param moves:
            A list of UCI moves (e.g. ['e2e4', 'e7e5', 'g1f3', 'b8f6'])
        """
        board = chess.Board(self.fen)
        for move in moves:
            board.push_uci(move)
        self.fen = board.fen()

    def apply_props(self):
        """
        Applies all props to the engine. Possible props may
        include: Hash table size, FEN position, openings database,
        etc.
        """
        self.cmd('uci')
        self.cmd(f'position fen {self.fen}')

    def best_move_uci(self, max_secs: float = 0, callback: Callable[[str], None] = None) -> str:
        """
        Finds the best move in the current position. The move is
        returned in UCI notation.

        :param max_secs:
            Maximum time to think about the current position

        :param callback:
            A function that is called each time the engine finds a potential
            'best move'. The function accepts a string as an argument (the move
            in UCI notation). Therefore, the last call of this function is called
            with the best move found.

        :return:
            The best move in UCI notation.
        """
        x = ''
        thinking = True

        # called after each line read
        def _call(line):
            m = re.findall('pv [a-h][1-8][a-h][1-8][qrnb]?', line)
            if len(m) > 0:
                nonlocal x
                u = m[-1][3:]
                x = u
                if callback is not None:
                    callback(u)

        # stop thinking after timer runs out
        def _timer():
            nonlocal thinking
            self.process.stdin.write('isready\n'.encode())
            self.process.stdin.flush()
            thinking = False

        self.apply_props()

        #######################
        # start finding moves #
        #######################

        self.process.stdin.write('go ponder\n'.encode())
        self.process.stdin.flush()
        threading.Timer(max_secs, _timer).start()

        while True:
            text = str(self.process.stdout.readline().strip())
            _call(text)
            if not thinking or 'bestmove' in text or 'readyok' in text:
                break

        ########################
        # stopped finding moves
        ########################

        self.cmd('stop')

        return x

    def best_move_san(self, max_secs: float = 0, callback: Callable[[str], None] = None) -> str:
        """
        Finds the best move in the current position. The move is
        returned in SAN notation. Note that this method a wrapper
        for Engine.best_move_uci().

        :param max_secs:
            Maximum time to think about the current position

        :param callback:
            A function that is called each time the engine finds a potential
            'best move'. The function accepts a string as an argument (the move
            in SAN notation). Therefore, the last call of this function is called
            with the best move found.

        :return:
            The best move in SAN notation.
        """

        def _call(m):
            if callback is not None:
                callback(chess.Board(self.fen).san(chess.Move.from_uci(m)))

        uci = self.best_move_uci(max_secs, callback=_call)
        board = chess.Board(self.fen)
        move = chess.Move.from_uci(uci)
        return board.san(move)


class Stockfish(Engine):
    """
    This class is an implementation of Stockfish.
    """

    def __init__(self, exe: str = None, fen: str = None, props: Dict[str, Any] = None):
        """
        Connects to the Stockfish executable. By default, Stockfish 14.1 is used,
        if no executable is specified.

        :param exe:
            The Stockfish executable (default: sf-v.14.1)

        :param fen:
            The FEN position of the board (default: starting position)

        :param props:
            Optional properties for Stockfish
        """
        if exe is None:
            exe = 'Engines/Stockfish/Stockfish-sf_14.1/src/stockfish'
        super().__init__(exe, fen, props)


class Leela(Engine):
    """
    This class is an implementation for Leela Chess Zero (lc0).
    """

    def __init__(self, exe: str = None, fen: str = None, props: Dict[str, Any] = None):
        """
        Connects to the Leela Chess Zero executable. If no executable is specified, a
        default version is used instead.

        :param exe:
            The lc0 executable

        :param fen:
            The FEN position of the board (default: starting position)

        :param props:
            Optionals properties for lc0
        """
        if exe is None:
            exe = 'Engines/Leela/lc0-master/build/release/lc0'
        super().__init__(exe, fen, props)



#  This file is part of Leela Chess Zero.
#  Copyright (C) 2020 The LCZero Authors
#
#  Leela Chess is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Leela Chess is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Leela Chess.  If not, see <http://www.gnu.org/licenses/>.
#
#  Additional permission under GNU GPL version 3 section 7
#
#  If you modify this Program, or any covered work, by linking or
#  combining it with NVIDIA Corporation's libraries from the NVIDIA CUDA
#  Toolkit and the NVIDIA CUDA Deep Neural Network library (or a
#  modified version of those libraries), containing parts covered by the
#  terms of the respective license agreement, the licensors of this
#  Program grant you additional permission to convey the resulting work.

from .core import PyObject


class CppException(PyObject):
    def __init__(self, name, cpp_name, *args, **kwargs):
        self.cpp_name = cpp_name
        super().__init__(name, *args, **kwargs)

    def type_struct_name(self):
        return f'T{self.name}ExceptionType'

    def Generate(self, w):
        w.Write(f'PyObject *{self.type_struct_name()};')

    def GenerateRegister(self, w, module):
        w.Write(f'{self.type_struct_name()} = PyErr_NewException('
                f'"{module.name}.{self.name}", nullptr, nullptr);')
        w.Write(f'if ({self.type_struct_name()} == nullptr) return nullptr;')
        w.Write(f'Py_INCREF({self.type_struct_name()});')
        w.Write(f'PyModule_AddObject(module, "{self.name}", '
                f'{self.type_struct_name()});')

    def GenerateHandle(self, w, func):
        w.Unindent()
        w.Open(f'}} catch (const {self.cpp_name} &ex) {{')
        w.Write(f'PyErr_SetString({self.type_struct_name()}, ex.what());')
        w.Write(f'return {func._failure()};')



# Generated from /home/mario/Documents/PYTHON/chessEngines/parser/gen/pgn.g4 by ANTLR 4.9.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\24")
        buf.write("\u00ac\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\3\2\3\2\3\2\3\2\7\2\27\n\2\f\2\16\2\32")
        buf.write("\13\2\3\2\3\2\3\2\3\3\3\3\3\3\3\3\7\3#\n\3\f\3\16\3&\13")
        buf.write("\3\3\3\3\3\3\3\3\4\3\4\3\4\3\4\6\4/\n\4\r\4\16\4\60\3")
        buf.write("\4\7\4\64\n\4\f\4\16\4\67\13\4\3\4\3\4\3\4\3\4\3\5\3\5")
        buf.write("\3\5\3\5\3\6\3\6\5\6C\n\6\3\6\3\6\3\7\3\7\3\7\3\7\3\7")
        buf.write("\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3")
        buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7`\n\7\3\7\5\7c\n\7\3\7\7")
        buf.write("\7f\n\7\f\7\16\7i\13\7\3\7\3\7\3\7\5\7n\n\7\3\7\3\7\3")
        buf.write("\b\3\b\3\b\3\b\7\bv\n\b\f\b\16\by\13\b\3\b\3\b\3\b\6\b")
        buf.write("~\n\b\r\b\16\b\177\3\b\3\b\3\b\5\b\u0085\n\b\3\b\3\b\3")
        buf.write("\b\6\b\u008a\n\b\r\b\16\b\u008b\3\b\7\b\u008f\n\b\f\b")
        buf.write("\16\b\u0092\13\b\3\b\3\b\3\b\5\b\u0097\n\b\3\b\3\b\3\t")
        buf.write("\3\t\7\t\u009d\n\t\f\t\16\t\u00a0\13\t\3\t\3\t\7\t\u00a4")
        buf.write("\n\t\f\t\16\t\u00a7\13\t\3\t\5\t\u00aa\n\t\3\t\b\65gw")
        buf.write("\u0090\u009e\u00a5\2\n\2\4\6\b\n\f\16\20\2\b\3\2\3\3\3")
        buf.write("\2\5\5\3\2\24\24\3\2\r\16\3\2\n\13\4\2\16\16\22\22\2\u00bb")
        buf.write("\2\22\3\2\2\2\4\36\3\2\2\2\6*\3\2\2\2\b<\3\2\2\2\n@\3")
        buf.write("\2\2\2\fF\3\2\2\2\16q\3\2\2\2\20\u009e\3\2\2\2\22\23\b")
        buf.write("\2\1\2\23\30\7\3\2\2\24\25\n\2\2\2\25\27\b\2\1\2\26\24")
        buf.write("\3\2\2\2\27\32\3\2\2\2\30\26\3\2\2\2\30\31\3\2\2\2\31")
        buf.write("\33\3\2\2\2\32\30\3\2\2\2\33\34\7\3\2\2\34\35\b\2\1\2")
        buf.write("\35\3\3\2\2\2\36\37\b\3\1\2\37$\7\4\2\2 !\n\3\2\2!#\b")
        buf.write("\3\1\2\" \3\2\2\2#&\3\2\2\2$\"\3\2\2\2$%\3\2\2\2%\'\3")
        buf.write("\2\2\2&$\3\2\2\2\'(\7\5\2\2()\b\3\1\2)\5\3\2\2\2*+\b\4")
        buf.write("\1\2+.\7\6\2\2,-\n\4\2\2-/\b\4\1\2.,\3\2\2\2/\60\3\2\2")
        buf.write("\2\60.\3\2\2\2\60\61\3\2\2\2\61\65\3\2\2\2\62\64\7\24")
        buf.write("\2\2\63\62\3\2\2\2\64\67\3\2\2\2\65\66\3\2\2\2\65\63\3")
        buf.write("\2\2\2\668\3\2\2\2\67\65\3\2\2\289\5\2\2\29:\b\4\1\2:")
        buf.write(";\7\7\2\2;\7\3\2\2\2<=\7\r\2\2=>\7\16\2\2>?\b\5\1\2?\t")
        buf.write("\3\2\2\2@B\7\17\2\2AC\t\5\2\2BA\3\2\2\2BC\3\2\2\2CD\3")
        buf.write("\2\2\2DE\b\6\1\2E\13\3\2\2\2F_\b\7\1\2GH\5\b\5\2HI\b\7")
        buf.write("\1\2I`\3\2\2\2JK\7\r\2\2KL\7\b\2\2LM\5\b\5\2MN\b\7\1\2")
        buf.write("N`\3\2\2\2OP\5\n\6\2PQ\7\b\2\2QR\5\b\5\2RS\b\7\1\2S`\3")
        buf.write("\2\2\2TU\5\n\6\2UV\5\b\5\2VW\b\7\1\2W`\3\2\2\2XY\5\b\5")
        buf.write("\2YZ\7\t\2\2Z[\5\n\6\2[\\\b\7\1\2\\`\3\2\2\2]^\t\6\2\2")
        buf.write("^`\b\7\1\2_G\3\2\2\2_J\3\2\2\2_O\3\2\2\2_T\3\2\2\2_X\3")
        buf.write("\2\2\2_]\3\2\2\2`b\3\2\2\2ac\7\20\2\2ba\3\2\2\2bc\3\2")
        buf.write("\2\2cg\3\2\2\2df\7\24\2\2ed\3\2\2\2fi\3\2\2\2gh\3\2\2")
        buf.write("\2ge\3\2\2\2hm\3\2\2\2ig\3\2\2\2jk\5\4\3\2kl\b\7\1\2l")
        buf.write("n\3\2\2\2mj\3\2\2\2mn\3\2\2\2no\3\2\2\2op\b\7\1\2p\r\3")
        buf.write("\2\2\2qr\b\b\1\2rs\t\7\2\2sw\7\f\2\2tv\7\24\2\2ut\3\2")
        buf.write("\2\2vy\3\2\2\2wx\3\2\2\2wu\3\2\2\2x\u0096\3\2\2\2yw\3")
        buf.write("\2\2\2z{\5\f\7\2{\u0084\b\b\1\2|~\7\24\2\2}|\3\2\2\2~")
        buf.write("\177\3\2\2\2\177}\3\2\2\2\177\u0080\3\2\2\2\u0080\u0081")
        buf.write("\3\2\2\2\u0081\u0082\5\f\7\2\u0082\u0083\b\b\1\2\u0083")
        buf.write("\u0085\3\2\2\2\u0084}\3\2\2\2\u0084\u0085\3\2\2\2\u0085")
        buf.write("\u0086\3\2\2\2\u0086\u0087\b\b\1\2\u0087\u0097\3\2\2\2")
        buf.write("\u0088\u008a\7\f\2\2\u0089\u0088\3\2\2\2\u008a\u008b\3")
        buf.write("\2\2\2\u008b\u0089\3\2\2\2\u008b\u008c\3\2\2\2\u008c\u0090")
        buf.write("\3\2\2\2\u008d\u008f\7\24\2\2\u008e\u008d\3\2\2\2\u008f")
        buf.write("\u0092\3\2\2\2\u0090\u0091\3\2\2\2\u0090\u008e\3\2\2\2")
        buf.write("\u0091\u0093\3\2\2\2\u0092\u0090\3\2\2\2\u0093\u0094\5")
        buf.write("\f\7\2\u0094\u0095\b\b\1\2\u0095\u0097\3\2\2\2\u0096z")
        buf.write("\3\2\2\2\u0096\u0089\3\2\2\2\u0097\u0098\3\2\2\2\u0098")
        buf.write("\u0099\b\b\1\2\u0099\17\3\2\2\2\u009a\u009d\5\6\4\2\u009b")
        buf.write("\u009d\7\24\2\2\u009c\u009a\3\2\2\2\u009c\u009b\3\2\2")
        buf.write("\2\u009d\u00a0\3\2\2\2\u009e\u009f\3\2\2\2\u009e\u009c")
        buf.write("\3\2\2\2\u009f\u00a5\3\2\2\2\u00a0\u009e\3\2\2\2\u00a1")
        buf.write("\u00a4\5\16\b\2\u00a2\u00a4\7\24\2\2\u00a3\u00a1\3\2\2")
        buf.write("\2\u00a3\u00a2\3\2\2\2\u00a4\u00a7\3\2\2\2\u00a5\u00a6")
        buf.write("\3\2\2\2\u00a5\u00a3\3\2\2\2\u00a6\u00a9\3\2\2\2\u00a7")
        buf.write("\u00a5\3\2\2\2\u00a8\u00aa\7\21\2\2\u00a9\u00a8\3\2\2")
        buf.write("\2\u00a9\u00aa\3\2\2\2\u00aa\21\3\2\2\2\26\30$\60\65B")
        buf.write("_bgmw\177\u0084\u008b\u0090\u0096\u009c\u009e\u00a3\u00a5")
        buf.write("\u00a9")
        return buf.getvalue()


class pgnParser ( Parser ):

    grammarFileName = "pgn.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'\"'", "'{'", "'}'", "'['", "']'", "'x'", 
                     "'='", "'O-O'", "'O-O-O'", "'.'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "FILE", "RANK", 
                      "PIECE", "GLYPH", "RESULT", "INT", "CHAR", "WS" ]

    RULE_string = 0
    RULE_comment = 1
    RULE_tag = 2
    RULE_square = 3
    RULE_piece = 4
    RULE_halfMove = 5
    RULE_fullMove = 6
    RULE_game = 7

    ruleNames =  [ "string", "comment", "tag", "square", "piece", "halfMove", 
                   "fullMove", "game" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    FILE=11
    RANK=12
    PIECE=13
    GLYPH=14
    RESULT=15
    INT=16
    CHAR=17
    WS=18

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.9.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    tags = {}
    moves = []



    class StringContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.txt = None
            self.c = None # Token


        def getRuleIndex(self):
            return pgnParser.RULE_string

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterString" ):
                listener.enterString(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitString" ):
                listener.exitString(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitString" ):
                return visitor.visitString(self)
            else:
                return visitor.visitChildren(self)




    def string(self):

        localctx = pgnParser.StringContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_string)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)

            arr = []

            self.state = 17
            self.match(pgnParser.T__0)
            self.state = 22
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << pgnParser.T__1) | (1 << pgnParser.T__2) | (1 << pgnParser.T__3) | (1 << pgnParser.T__4) | (1 << pgnParser.T__5) | (1 << pgnParser.T__6) | (1 << pgnParser.T__7) | (1 << pgnParser.T__8) | (1 << pgnParser.T__9) | (1 << pgnParser.FILE) | (1 << pgnParser.RANK) | (1 << pgnParser.PIECE) | (1 << pgnParser.GLYPH) | (1 << pgnParser.RESULT) | (1 << pgnParser.INT) | (1 << pgnParser.CHAR) | (1 << pgnParser.WS))) != 0):
                self.state = 18
                localctx.c = self._input.LT(1)
                _la = self._input.LA(1)
                if _la <= 0 or _la==pgnParser.T__0:
                    localctx.c = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()

                arr.append((None if localctx.c is None else localctx.c.text))

                self.state = 24
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 25
            self.match(pgnParser.T__0)

            localctx.txt = ''.join(arr)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CommentContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.txt = None
            self.c = None # Token


        def getRuleIndex(self):
            return pgnParser.RULE_comment

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComment" ):
                listener.enterComment(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComment" ):
                listener.exitComment(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitComment" ):
                return visitor.visitComment(self)
            else:
                return visitor.visitChildren(self)




    def comment(self):

        localctx = pgnParser.CommentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_comment)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)

            arr = []

            self.state = 29
            self.match(pgnParser.T__1)
            self.state = 34
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << pgnParser.T__0) | (1 << pgnParser.T__1) | (1 << pgnParser.T__3) | (1 << pgnParser.T__4) | (1 << pgnParser.T__5) | (1 << pgnParser.T__6) | (1 << pgnParser.T__7) | (1 << pgnParser.T__8) | (1 << pgnParser.T__9) | (1 << pgnParser.FILE) | (1 << pgnParser.RANK) | (1 << pgnParser.PIECE) | (1 << pgnParser.GLYPH) | (1 << pgnParser.RESULT) | (1 << pgnParser.INT) | (1 << pgnParser.CHAR) | (1 << pgnParser.WS))) != 0):
                self.state = 30
                localctx.c = self._input.LT(1)
                _la = self._input.LA(1)
                if _la <= 0 or _la==pgnParser.T__2:
                    localctx.c = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()

                arr.append((None if localctx.c is None else localctx.c.text))

                self.state = 36
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 37
            self.match(pgnParser.T__2)

            localctx.txt = ''.join(arr)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class TagContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.a = None # Token
            self.s = None # StringContext

        def string(self):
            return self.getTypedRuleContext(pgnParser.StringContext,0)


        def WS(self, i:int=None):
            if i is None:
                return self.getTokens(pgnParser.WS)
            else:
                return self.getToken(pgnParser.WS, i)

        def getRuleIndex(self):
            return pgnParser.RULE_tag

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterTag" ):
                listener.enterTag(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitTag" ):
                listener.exitTag(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitTag" ):
                return visitor.visitTag(self)
            else:
                return visitor.visitChildren(self)




    def tag(self):

        localctx = pgnParser.TagContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_tag)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)

            arr = []
            key = ''

            self.state = 41
            self.match(pgnParser.T__3)
            self.state = 44 
            self._errHandler.sync(self)
            _alt = 1
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 42
                    localctx.a = self._input.LT(1)
                    _la = self._input.LA(1)
                    if _la <= 0 or _la==pgnParser.WS:
                        localctx.a = self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()

                    arr.append((None if localctx.a is None else localctx.a.text))


                else:
                    raise NoViableAltException(self)
                self.state = 46 
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,2,self._ctx)

            self.state = 51
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,3,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 48
                    self.match(pgnParser.WS) 
                self.state = 53
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,3,self._ctx)

            self.state = 54
            localctx.s = self.string()

            key = ''.join(arr)
            val = (None if localctx.s is None else self._input.getText(localctx.s.start,localctx.s.stop))[1:-1]
            self.tags[key] = val;

            self.state = 56
            self.match(pgnParser.T__4)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class SquareContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.txt = None
            self.f = None # Token
            self.r = None # Token

        def FILE(self):
            return self.getToken(pgnParser.FILE, 0)

        def RANK(self):
            return self.getToken(pgnParser.RANK, 0)

        def getRuleIndex(self):
            return pgnParser.RULE_square

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSquare" ):
                listener.enterSquare(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSquare" ):
                listener.exitSquare(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSquare" ):
                return visitor.visitSquare(self)
            else:
                return visitor.visitChildren(self)




    def square(self):

        localctx = pgnParser.SquareContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_square)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 58
            localctx.f = self.match(pgnParser.FILE)
            self.state = 59
            localctx.r = self.match(pgnParser.RANK)

            localctx.txt = (None if localctx.f is None else localctx.f.text) + (None if localctx.r is None else localctx.r.text)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PieceContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.txt = None
            self.p = None # Token
            self.x = None # Token

        def PIECE(self):
            return self.getToken(pgnParser.PIECE, 0)

        def FILE(self):
            return self.getToken(pgnParser.FILE, 0)

        def RANK(self):
            return self.getToken(pgnParser.RANK, 0)

        def getRuleIndex(self):
            return pgnParser.RULE_piece

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPiece" ):
                listener.enterPiece(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPiece" ):
                listener.exitPiece(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPiece" ):
                return visitor.visitPiece(self)
            else:
                return visitor.visitChildren(self)




    def piece(self):

        localctx = pgnParser.PieceContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_piece)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 62
            localctx.p = self.match(pgnParser.PIECE)
            self.state = 64
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,4,self._ctx)
            if la_ == 1:
                self.state = 63
                localctx.x = self._input.LT(1)
                _la = self._input.LA(1)
                if not(_la==pgnParser.FILE or _la==pgnParser.RANK):
                    localctx.x = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()



            localctx.txt = (None if localctx.p is None else localctx.p.text)
            if localctx.x is not None:
                localctx.txt += (None if localctx.x is None else localctx.x.text)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class HalfMoveContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.props = None
            self.s = None # SquareContext
            self.f = None # Token
            self.p = None # PieceContext
            self.x = None # Token
            self.g = None # Token
            self.c = None # CommentContext

        def square(self):
            return self.getTypedRuleContext(pgnParser.SquareContext,0)


        def FILE(self):
            return self.getToken(pgnParser.FILE, 0)

        def piece(self):
            return self.getTypedRuleContext(pgnParser.PieceContext,0)


        def WS(self, i:int=None):
            if i is None:
                return self.getTokens(pgnParser.WS)
            else:
                return self.getToken(pgnParser.WS, i)

        def GLYPH(self):
            return self.getToken(pgnParser.GLYPH, 0)

        def comment(self):
            return self.getTypedRuleContext(pgnParser.CommentContext,0)


        def getRuleIndex(self):
            return pgnParser.RULE_halfMove

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterHalfMove" ):
                listener.enterHalfMove(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitHalfMove" ):
                listener.exitHalfMove(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitHalfMove" ):
                return visitor.visitHalfMove(self)
            else:
                return visitor.visitChildren(self)




    def halfMove(self):

        localctx = pgnParser.HalfMoveContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_halfMove)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)

            localctx.props = {
                'move': None,
                'glyph': None,
                'comment': None
            }

            self.state = 93
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,5,self._ctx)
            if la_ == 1:
                self.state = 69
                localctx.s = self.square()

                localctx.props['move'] = localctx.s.txt

                pass

            elif la_ == 2:
                self.state = 72
                localctx.f = self.match(pgnParser.FILE)
                self.state = 73
                self.match(pgnParser.T__5)
                self.state = 74
                localctx.s = self.square()

                localctx.props['move'] = (None if localctx.f is None else localctx.f.text) + 'x' + localctx.s.txt

                pass

            elif la_ == 3:
                self.state = 77
                localctx.p = self.piece()
                self.state = 78
                self.match(pgnParser.T__5)
                self.state = 79
                localctx.s = self.square()

                localctx.props['move'] = localctx.p.txt + 'x' + localctx.s.txt

                pass

            elif la_ == 4:
                self.state = 82
                localctx.p = self.piece()
                self.state = 83
                localctx.s = self.square()

                localctx.props['move'] = localctx.p.txt + localctx.s.txt

                pass

            elif la_ == 5:
                self.state = 86
                localctx.s = self.square()
                self.state = 87
                self.match(pgnParser.T__6)
                self.state = 88
                localctx.p = self.piece()

                localctx.props['move'] = localctx.s.txt + '=' + localctx.p.txt

                pass

            elif la_ == 6:
                self.state = 91
                localctx.x = self._input.LT(1)
                _la = self._input.LA(1)
                if not(_la==pgnParser.T__7 or _la==pgnParser.T__8):
                    localctx.x = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()

                localctx.props['move'] = (None if localctx.x is None else localctx.x.text)

                pass


            self.state = 96
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==pgnParser.GLYPH:
                self.state = 95
                localctx.g = self.match(pgnParser.GLYPH)


            self.state = 101
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,7,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 98
                    self.match(pgnParser.WS) 
                self.state = 103
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,7,self._ctx)

            self.state = 107
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==pgnParser.T__1:
                self.state = 104
                localctx.c = self.comment()

                localctx.props['comment'] = localctx.c.txt




            localctx.props['glyph'] = None if (localctx.g is None) else (None if localctx.g is None else localctx.g.text)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class FullMoveContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.h1 = None # HalfMoveContext
            self.h2 = None # HalfMoveContext
            self.h = None # HalfMoveContext

        def INT(self):
            return self.getToken(pgnParser.INT, 0)

        def RANK(self):
            return self.getToken(pgnParser.RANK, 0)

        def WS(self, i:int=None):
            if i is None:
                return self.getTokens(pgnParser.WS)
            else:
                return self.getToken(pgnParser.WS, i)

        def halfMove(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(pgnParser.HalfMoveContext)
            else:
                return self.getTypedRuleContext(pgnParser.HalfMoveContext,i)


        def getRuleIndex(self):
            return pgnParser.RULE_fullMove

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFullMove" ):
                listener.enterFullMove(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFullMove" ):
                listener.exitFullMove(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitFullMove" ):
                return visitor.visitFullMove(self)
            else:
                return visitor.visitChildren(self)




    def fullMove(self):

        localctx = pgnParser.FullMoveContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_fullMove)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)

            move = []

            self.state = 112
            _la = self._input.LA(1)
            if not(_la==pgnParser.RANK or _la==pgnParser.INT):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 113
            self.match(pgnParser.T__9)
            self.state = 117
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,9,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 114
                    self.match(pgnParser.WS) 
                self.state = 119
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,9,self._ctx)

            self.state = 148
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [pgnParser.T__7, pgnParser.T__8, pgnParser.FILE, pgnParser.PIECE]:
                self.state = 120
                localctx.h1 = self.halfMove()

                move.append(localctx.h1.props)

                self.state = 130
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,11,self._ctx)
                if la_ == 1:
                    self.state = 123 
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    while True:
                        self.state = 122
                        self.match(pgnParser.WS)
                        self.state = 125 
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)
                        if not (_la==pgnParser.WS):
                            break

                    self.state = 127
                    localctx.h2 = self.halfMove()

                    move.append(localctx.h2.props)




                if len(move) < 2:
                    move.append({
                        'move': None,
                        'glyph': None,
                        'comment': None
                    })

                pass
            elif token in [pgnParser.T__9]:
                self.state = 135 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 134
                    self.match(pgnParser.T__9)
                    self.state = 137 
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la==pgnParser.T__9):
                        break

                self.state = 142
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,13,self._ctx)
                while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt==1+1:
                        self.state = 139
                        self.match(pgnParser.WS) 
                    self.state = 144
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,13,self._ctx)

                self.state = 145
                localctx.h = self.halfMove()

                move.append({
                    'move': None,
                    'glyph': None,
                    'comment': None
                })
                move.append(localctx.h.props)

                pass
            else:
                raise NoViableAltException(self)


            self.moves.append(move)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class GameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def tag(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(pgnParser.TagContext)
            else:
                return self.getTypedRuleContext(pgnParser.TagContext,i)


        def WS(self, i:int=None):
            if i is None:
                return self.getTokens(pgnParser.WS)
            else:
                return self.getToken(pgnParser.WS, i)

        def fullMove(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(pgnParser.FullMoveContext)
            else:
                return self.getTypedRuleContext(pgnParser.FullMoveContext,i)


        def RESULT(self):
            return self.getToken(pgnParser.RESULT, 0)

        def getRuleIndex(self):
            return pgnParser.RULE_game

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterGame" ):
                listener.enterGame(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitGame" ):
                listener.exitGame(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitGame" ):
                return visitor.visitGame(self)
            else:
                return visitor.visitChildren(self)




    def game(self):

        localctx = pgnParser.GameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_game)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 156
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,16,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 154
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [pgnParser.T__3]:
                        self.state = 152
                        self.tag()
                        pass
                    elif token in [pgnParser.WS]:
                        self.state = 153
                        self.match(pgnParser.WS)
                        pass
                    else:
                        raise NoViableAltException(self)
             
                self.state = 158
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,16,self._ctx)

            self.state = 163
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,18,self._ctx)
            while _alt!=1 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1+1:
                    self.state = 161
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [pgnParser.RANK, pgnParser.INT]:
                        self.state = 159
                        self.fullMove()
                        pass
                    elif token in [pgnParser.WS]:
                        self.state = 160
                        self.match(pgnParser.WS)
                        pass
                    else:
                        raise NoViableAltException(self)
             
                self.state = 165
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,18,self._ctx)

            self.state = 167
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==pgnParser.RESULT:
                self.state = 166
                self.match(pgnParser.RESULT)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








#!/usr/bin/env python
# Call dxc with absolute paths where '/' are converted to '\' (in windows).
import sys
import os
import re

x = sys.argv
x.pop(0)
for i in range(len(x)):
  if re.match(r"[a-zA-Z]:/", x[i]):
    x[i] = os.path.normpath(x[i])
# We asssume dxc is already on the path.
os.system('dxc ' + '"{}"'.format('" "'.join(x)))




#  This file is part of Leela Chess Zero.
#  Copyright (C) 2020 The LCZero Authors
#
#  Leela Chess is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Leela Chess is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Leela Chess.  If not, see <http://www.gnu.org/licenses/>.
#
#  Additional permission under GNU GPL version 3 section 7
#
#  If you modify this Program, or any covered work, by linking or
#  combining it with NVIDIA Corporation's libraries from the NVIDIA CUDA
#  Toolkit and the NVIDIA CUDA Deep Neural Network library (or a
#  modified version of those libraries), containing parts covered by the
#  terms of the respective license agreement, the licensors of this
#  Program grant you additional permission to convey the resulting work.

from .core import PyObject
from .functions import (Constructor, MemberFunction, StaticFunction,
                        DisabledConstructor)
from .exceptions import (CppException)


class FunctionContainer(PyObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.functions = []

    def _generate_functions(self, w):
        for func in self.functions:
            func.Generate(w)
        self._generate_function_list(w)

    def _generate_function_list(self, w):
        w.Open(f'PyMethodDef {self.function_list_name()}[] = {{')
        for f in self.functions:
            w.Write(f'{{"{f.name}", '
                    f'reinterpret_cast<PyCFunction>(&{f.gen_function_name}), '
                    f'{f.function_meth_flags()}, {f.BuildDocString()}}},')
        w.Write('{nullptr, nullptr, 0, nullptr}')
        w.Close('};\n')


class Class(FunctionContainer):
    def __init__(self,
                 name,
                 *argv,
                 cpp_name=None,
                 disable_constructor=False,
                 **kwargs):
        self.cpp_name = cpp_name or name
        super().__init__(name, *argv, **kwargs)
        if disable_constructor:
            self.constructor = DisabledConstructor(
                gen_function_name=self.constructor_name())
        else:
            self.constructor = Constructor(
                cpp_type_name=self.cpp_name,
                gen_function_name=self.constructor_name(),
                self_type=self.object_struct_name())

    def GenerateForwardDeclarations(self, w):
        w.Write(f'struct {self.object_struct_name()};')
        w.Write(f'extern PyTypeObject {self.type_object_name()};')

    def Generate(self, w, module):
        # Object type.
        w.Open(f'struct {self.object_struct_name()} {{')
        w.Write('PyObject_HEAD')
        w.Write(f'{self.cpp_name} *value;')
        w.Close('};\n')

        # Functions
        self._generate_functions(w)

        # Constructor and destructor.
        self.constructor.Generate(w)
        self._generate_destructor(w)

        # Type object.
        self._generate_class_struct(w, module)

    def GenerateRegister(self, w):
        w.Write(f'if (PyType_Ready(&{self.type_object_name()}) '
                '!= 0) return nullptr;')
        w.Write(f'PyModule_AddObject(module, "{self.name}", '
                f'&{self.type_object_name()}.ob_base.ob_base);')

    def AddMethod(self, name, cpp_name=None, *argv, **kwargs):
        method = MemberFunction(name,
                                cpp_name=cpp_name,
                                self_type=self.object_struct_name(),
                                gen_function_name=f'F{self.name}Method{name}',
                                *argv,
                                **kwargs)
        self.functions.append(method)
        return method

    def AddStaticMethod(self, name, cpp_name=None, *argv, **kwargs):
        method = StaticFunction(
            name,
            cpp_type_name=self.cpp_name,
            cpp_name=cpp_name,
            gen_function_name=f'F{self.name}StaticMethod{name}',
            *argv,
            **kwargs)
        self.functions.append(method)
        return method

    def object_struct_name(self):
        return f'T{self.name}ClassType'

    def type_object_name(self):
        return f'obj{self.name}ClassType'

    def constructor_name(self):
        return f'F{self.name}Constructor'

    def destructor_name(self):
        return f'F{self.name}Destructor'

    def function_list_name(self):
        return f'rg{self.name}ClassFunctions'

    def _generate_class_struct(self, w, module):
        w.Open(f'PyTypeObject {self.type_object_name()} = {{')
        w.Write('.ob_base = PyVarObject_HEAD_INIT(NULL, 0)')
        w.Write(f'.tp_name = "{module.name}.{self.name}",')
        w.Write(f'.tp_basicsize = sizeof({self.object_struct_name()}),')
        w.Write('.tp_dealloc = reinterpret_cast<destructor>'
                f'({self.destructor_name()}),')
        w.Write('.tp_flags = Py_TPFLAGS_DEFAULT,')
        w.Write(f'.tp_doc = {self.BuildDocString()},')
        w.Write(f'.tp_methods = {self.function_list_name()},')
        w.Write('.tp_init = reinterpret_cast<initproc>'
                f'({self.constructor_name()}),')
        w.Write('.tp_alloc = PyType_GenericAlloc,')
        w.Write('.tp_new = PyType_GenericNew,')
        w.Close('};')

    def _generate_destructor(self, w):
        w.Open(f'void {self.destructor_name()}('
               f'{self.object_struct_name()}* self) {{')
        w.Write('delete self->value;')
        w.Write('Py_TYPE(self)->tp_free(&self->ob_base);')
        w.Close('}\n')


class Module(FunctionContainer):
    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.includes = []
        self.classes = []
        self.exceptions = []
        self.initialization = []

    def AddInclude(self, s):
        assert isinstance(s, str)
        self.includes.append(s)

    def AddClass(self, cls):
        assert isinstance(cls, Class)
        self.classes.append(cls)
        return cls

    def AddException(self, ex):
        assert isinstance(ex, CppException)
        self.exceptions.append(ex)
        return ex

    def AddInitialization(self, s):
        assert isinstance(s, str)
        self.initialization.append(s)

    def Generate(self, w):
        w.Write('// This file is AUTOGENERATED, do not edit.')
        w.Write('#define PY_SSIZE_T_CLEAN')
        w.Write('#include <Python.h>')
        for x in self.includes:
            w.Write(f'#include "{x}"')

        w.Write('\nnamespace {')
        for cls in self.exceptions:
            cls.Generate(w)
        for cls in self.classes:
            cls.GenerateForwardDeclarations(w)
        for cls in self.classes:
            cls.Generate(w, self)
        self._generate_functions(w)
        self._generate_module_stuct(w)
        w.Write('}  // anonymous namespace\n')

        self._generate_main_func(w)

    def struct_name(self):
        return f'T{self.name}Module'

    def function_list_name(self):
        return f'rg{self.name}ModuleFunctions'

    def _generate_module_stuct(self, w):
        w.Open(f'PyModuleDef {self.struct_name()} = {{')
        w.Write('PyModuleDef_HEAD_INIT,')
        w.Write(f'"{self.name}",')
        w.Write(f'{self.BuildDocString()},')
        w.Write('-1,')
        w.Write(f'{self.function_list_name()},')
        w.Write('nullptr, ' * 4)
        w.Close('};')

    def _generate_main_func(self, w):
        w.Open(f'PyMODINIT_FUNC PyInit_{self.name}() {{')
        for x in self.initialization:
            w.Write('%s\n' % x)
        w.Write(f'PyObject* module = PyModule_Create(&{self.struct_name()});')
        w.Write('if (module == nullptr) return nullptr;')
        for x in self.exceptions:
            x.GenerateRegister(w, self)
        for x in self.classes:
            x.GenerateRegister(w)
        w.Write('return module;')
        w.Close('}')



#  This file is part of Leela Chess Zero.
#  Copyright (C) 2020 The LCZero Authors
#
#  Leela Chess is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Leela Chess is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Leela Chess.  If not, see <http://www.gnu.org/licenses/>.
#
#  Additional permission under GNU GPL version 3 section 7
#
#  If you modify this Program, or any covered work, by linking or
#  combining it with NVIDIA Corporation's libraries from the NVIDIA CUDA
#  Toolkit and the NVIDIA CUDA Deep Neural Network library (or a
#  modified version of those libraries), containing parts covered by the
#  terms of the respective license agreement, the licensors of this
#  Program grant you additional permission to convey the resulting work.


class RetVal:
    def cpp_val(self):
        return 'retval_cpp'

    def py_val(self):
        return 'retval'

    def ret_val(self):
        return self.py_val()

    def GenerateDeclaration(self, w):
        w.Write(f'PyObject *{self.py_val()};')

    def GenerateConversion(self, w):
        raise NotImplementedError()


class NoneRetVal(RetVal):
    def GenerateDeclaration(self, w):
        pass

    def GenerateConversion(self, w):
        pass

    def py_val(self):
        return 'Py_BuildValue("")'


class GenericStringRetVal(RetVal):
    def GenerateConversion(self, w):
        w.Write(f'{self.py_val()} = Py_BuildValue("s#", '
                f'{self.cpp_val()}.data(), {self.cpp_val()}.size());')


class StringViewRetVal(GenericStringRetVal):
    def cpp_type(self):
        return 'std::string_view'

    def GenerateConversion(self, w):
        w.Write(f'{self.py_val()} = Py_BuildValue("s#", '
                f'{self.cpp_val()}.data(), {self.cpp_val()}.size());')


class StringRetVal(GenericStringRetVal):
    def cpp_type(self):
        return 'const std::string&'

    def GenerateConversion(self, w):
        w.Write(f'{self.py_val()} = Py_BuildValue("s#", '
                f'{self.cpp_val()}.data(), {self.cpp_val()}.size());')


class ListOfStringsRetVal(RetVal):
    def cpp_type(self):
        return 'std::vector<std::string>'

    def GenerateConversion(self, w):
        w.Write(f'{self.py_val()} = PyList_New({self.cpp_val()}.size());')
        w.Open(f'for (size_t i = 0; i < {self.cpp_val()}.size(); ++i) {{')
        w.Write(f'const std::string& s = {self.cpp_val()}[i];')
        w.Write(f'PyList_SetItem({self.py_val()}, '
                'i, Py_BuildValue("s#", s.data(), s.size()));')
        w.Close('}')


class NumericRetVal(RetVal):
    def __init__(self, type):
        self.type = type

    def cpp_type(self):
        return {
            'i': 'int',
            'u64': 'uint64_t',
            'f32': 'float',
        }[self.type]

    def parse_tuple_format(self):
        return {
            'i': 'i',
            'u64': 'k',
            'f32': 'f',
        }[self.type]

    def GenerateConversion(self, w):
        w.Write(f'{self.py_val()} = Py_BuildValue('
                f'"{self.parse_tuple_format()}", {self.cpp_val()});')


class ObjCopyRetval(RetVal):
    def __init__(self, type):
        self.type = type

    def cpp_type(self):
        return f'const {self.type.cpp_name}&'

    def ret_val(self):
        return f'&{self.py_val()}->ob_base'

    def GenerateDeclaration(self, w):
        w.Write(f'{self.type.object_struct_name()} *{self.py_val()};')

    def GenerateConversion(self, w):
        w.Write(f'{self.py_val()} = PyObject_New('
                f'{self.type.object_struct_name()}, '
                f'&{self.type.type_object_name()});')
        w.Write(f'{self.py_val()}->value = '
                f'new {self.type.cpp_name}({self.cpp_val()});')


class ObjOwnerRetval(RetVal):
    def __init__(self, type):
        self.type = type

    def cpp_type(self):
        return f'std::unique_ptr<{self.type.cpp_name}>'

    def ret_val(self):
        return f'&{self.py_val()}->ob_base'

    def GenerateDeclaration(self, w):
        w.Write(f'{self.type.object_struct_name()} *{self.py_val()};')

    def GenerateConversion(self, w):
        w.Write(f'{self.py_val()} = PyObject_New('
                f'{self.type.object_struct_name()}, '
                f'&{self.type.type_object_name()});')
        w.Write(f'{self.py_val()}->value = ' f'{self.cpp_val()}.release();')


class ObjTupleRetVal(RetVal):
    def __init__(self, type):
        self.type = type

    def cpp_type(self):
        return f'std::vector<std::unique_ptr<{self.type.cpp_name}>>'

    def GenerateConversion(self, w):
        w.Write(f'{self.py_val()} = PyTuple_New({self.cpp_val()}.size());')
        w.Open(f'for (size_t i = 0; i < {self.cpp_val()}.size(); ++i) {{')
        w.Write(f'{self.type.object_struct_name()}* tmp = PyObject_New('
                f'{self.type.object_struct_name()}, '
                f'&{self.type.type_object_name()});')
        w.Write(f'tmp->value = {self.cpp_val()}[i].release();')
        w.Write(f'PyTuple_SetItem({self.py_val()}, i, &tmp->ob_base);')
        w.Close('}')


class IntegralTupleRetVal(RetVal):
    def __init__(self, type):
        self.type = type

    def cpp_item_type(self):
        return {
            'i': 'int',
            'u64': 'uint64_t',
            'f32': 'float',
        }[self.type]

    def parse_tuple_format(self):
        return {
            'i': 'i',
            'u64': 'k',
            'f32': 'f',
        }[self.type]

    def cpp_type(self):
        return f'std::vector<{self.cpp_item_type()}>'

    def GenerateConversion(self, w):
        w.Write(f'{self.py_val()} = PyTuple_New({self.cpp_val()}.size());')
        w.Open(f'for (size_t i = 0; i < {self.cpp_val()}.size(); ++i) {{')
        w.Write(f'PyTuple_SetItem({self.py_val()}, i, Py_BuildValue('
                f'"{self.parse_tuple_format()}", {self.cpp_val()}[i]));')
        w.Close('}')



# Generated from /home/mario/Documents/PYTHON/chessEngines/parser/gen/pgn.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .pgnParser import pgnParser
else:
    from pgnParser import pgnParser

# This class defines a complete listener for a parse tree produced by pgnParser.
class pgnListener(ParseTreeListener):

    # Enter a parse tree produced by pgnParser#string.
    def enterString(self, ctx:pgnParser.StringContext):
        pass

    # Exit a parse tree produced by pgnParser#string.
    def exitString(self, ctx:pgnParser.StringContext):
        pass


    # Enter a parse tree produced by pgnParser#comment.
    def enterComment(self, ctx:pgnParser.CommentContext):
        pass

    # Exit a parse tree produced by pgnParser#comment.
    def exitComment(self, ctx:pgnParser.CommentContext):
        pass


    # Enter a parse tree produced by pgnParser#tag.
    def enterTag(self, ctx:pgnParser.TagContext):
        pass

    # Exit a parse tree produced by pgnParser#tag.
    def exitTag(self, ctx:pgnParser.TagContext):
        pass


    # Enter a parse tree produced by pgnParser#square.
    def enterSquare(self, ctx:pgnParser.SquareContext):
        pass

    # Exit a parse tree produced by pgnParser#square.
    def exitSquare(self, ctx:pgnParser.SquareContext):
        pass


    # Enter a parse tree produced by pgnParser#piece.
    def enterPiece(self, ctx:pgnParser.PieceContext):
        pass

    # Exit a parse tree produced by pgnParser#piece.
    def exitPiece(self, ctx:pgnParser.PieceContext):
        pass


    # Enter a parse tree produced by pgnParser#halfMove.
    def enterHalfMove(self, ctx:pgnParser.HalfMoveContext):
        pass

    # Exit a parse tree produced by pgnParser#halfMove.
    def exitHalfMove(self, ctx:pgnParser.HalfMoveContext):
        pass


    # Enter a parse tree produced by pgnParser#fullMove.
    def enterFullMove(self, ctx:pgnParser.FullMoveContext):
        pass

    # Exit a parse tree produced by pgnParser#fullMove.
    def exitFullMove(self, ctx:pgnParser.FullMoveContext):
        pass


    # Enter a parse tree produced by pgnParser#game.
    def enterGame(self, ctx:pgnParser.GameContext):
        pass

    # Exit a parse tree produced by pgnParser#game.
    def exitGame(self, ctx:pgnParser.GameContext):
        pass



del pgnParser


import pathlib
from setuptools import setup, find_packages

CWD = pathlib.Path(__file__).parent
README = (CWD / "README.md").read_text()
REQUIREMENTS = [
    'antlr4-python3-runtime==4.9.3',
    'chess==1.7.0'
]

setup(
    name="chessEngines",
    version="1.0.8",
    author="Mario Di Caprio",
    author_email="mariodicaprio10@gmail.com",
    description="A wrapper for chess engines supporting the UCI protocol",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/MarioDiCaprio/chessEngines",
    project_urls={
        "Bug Tracker": "https://github.com/MarioDiCaprio/chessEngines/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=REQUIREMENTS
)



# Generated from /home/mario/Documents/PYTHON/chessEngines/parser/gen/pgn.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .pgnParser import pgnParser
else:
    from pgnParser import pgnParser

# This class defines a complete generic visitor for a parse tree produced by pgnParser.

class pgnVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by pgnParser#string.
    def visitString(self, ctx:pgnParser.StringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by pgnParser#comment.
    def visitComment(self, ctx:pgnParser.CommentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by pgnParser#tag.
    def visitTag(self, ctx:pgnParser.TagContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by pgnParser#square.
    def visitSquare(self, ctx:pgnParser.SquareContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by pgnParser#piece.
    def visitPiece(self, ctx:pgnParser.PieceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by pgnParser#halfMove.
    def visitHalfMove(self, ctx:pgnParser.HalfMoveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by pgnParser#fullMove.
    def visitFullMove(self, ctx:pgnParser.FullMoveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by pgnParser#game.
    def visitGame(self, ctx:pgnParser.GameContext):
        return self.visitChildren(ctx)



del pgnParser


#  This file is part of Leela Chess Zero.
#  Copyright (C) 2020 The LCZero Authors
#
#  Leela Chess is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Leela Chess is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Leela Chess.  If not, see <http://www.gnu.org/licenses/>.
#
#  Additional permission under GNU GPL version 3 section 7
#
#  If you modify this Program, or any covered work, by linking or
#  combining it with NVIDIA Corporation's libraries from the NVIDIA CUDA
#  Toolkit and the NVIDIA CUDA Deep Neural Network library (or a
#  modified version of those libraries), containing parts covered by the
#  terms of the respective license agreement, the licensors of this
#  Program grant you additional permission to convey the resulting work.


class Parameter:
    def __init__(self,
                 name,
                 optional=False,
                 cpp_name=None,
                 only_keyword=False):
        self.name = name
        self.cpp_name = cpp_name or name
        self.optional = optional
        self.only_keyword = only_keyword

    def name_at_caller(self):
        return self.cpp_name

    def needs_entire_argv(self):
        return False


class StringParameter(Parameter):
    def __init__(self, *argv, can_be_none=False, **kwargs):
        super().__init__(*argv, **kwargs)
        self.can_be_none = can_be_none

    def GenerateParseTupleSinkDeclaration(self, w):
        w.Write(f'const char* {self.name} = nullptr;')
        w.Write(f'Py_ssize_t {self.name}_len = 0;')

    def GenerateCppParamInitialization(self, w, func):
        if self.optional or self.can_be_none:
            w.Write(f'std::optional<std::string> {self.name_at_caller()};')
            w.Write(f'if ({self.name} != nullptr) '
                    f'{self.name_at_caller()}.emplace('
                    f'{self.name}, {self.name}_len);')
        else:
            raise NotImplementedError()

    def parse_tuple_sink_list(self):
        return [f'&{self.name}', f'&{self.name}_len']

    def parse_tuple_format(self):
        return 'z#' if self.can_be_none else 's#'

    def name_at_caller(self):
        return f'{self.cpp_name}_cpp'


class ClassParameter(Parameter):
    def __init__(self, cls, *args, **kwargs):
        self.cls = cls
        super().__init__(*args, **kwargs)

    def GenerateParseTupleSinkDeclaration(self, w):
        w.Write(f'{self.cls.object_struct_name()}* {self.name} = nullptr;')

    def parse_tuple_sink_list(self):
        return [f'&{self.cls.type_object_name()}', f'&{self.name}']

    def parse_tuple_format(self):
        return 'O!'

    def GenerateCppParamInitialization(self, w, func):
        pass

    def name_at_caller(self):
        if self.optional:
            return f'{self.cpp_name} ? {self.cpp_name}->value : nullptr'
        else:
            return f'*{self.cpp_name}->value'


class ListOfStringsParameter(Parameter):
    def GenerateParseTupleSinkDeclaration(self, w):
        w.Write(f'PyObject* {self.name} = nullptr;')

    def parse_tuple_format(self):
        return 'O!'

    def parse_tuple_sink_list(self):
        return ['&PyList_Type', f'&{self.name}']

    def GenerateCppParamInitialization(self, w, func):
        w.Write(f'std::vector<std::string> {self.name_at_caller()};')
        w.Open(f'if ({self.name} != nullptr) {{')
        w.Write(f'{self.name_at_caller()}.reserve(PyList_Size({self.name}));')
        w.Open(f'for (Py_ssize_t i = 0; i < PyList_Size({self.name}); ++i) {{')
        w.Write(f'PyObject* tmp = PyList_GetItem({self.name}, i);')
        w.Open('if (!PyUnicode_Check(tmp)) {')
        w.Write('PyErr_SetString(PyExc_TypeError, "String type expected.");')
        w.Write(f'return {func._failure()};')
        w.Close('}')
        w.Write('Py_ssize_t size;')
        w.Write('const char* str = PyUnicode_AsUTF8AndSize(tmp, &size);')
        w.Write(f'{self.name_at_caller()}.emplace_back(str, size);')
        w.Close('}')
        w.Close('}')

    def name_at_caller(self):
        return f'{self.cpp_name}_cpp'


class NumericParameter(Parameter):
    def __init__(self, *args, type='i', **kwargs):
        self.type = type
        super().__init__(*args, **kwargs)

    def GenerateParseTupleSinkDeclaration(self, w):
        w.Write(f'{self.cpp_type()} {self.name} = 0;')

    def cpp_type(self):
        return {
            'i': 'int',
            'u64': 'uint64_t',
            'f32': 'float',
        }[self.type]

    def parse_tuple_sink_list(self):
        return [f'&{self.name}']

    def parse_tuple_format(self):
        return {
            'i': 'i',
            'u64': 'k',
            'f32': 'f',
        }[self.type]

    def GenerateCppParamInitialization(self, w, func):
        pass


class ArgvParameter(Parameter):
    def __init__(self, name, type, *argv, **kwargs):
        self.type = type
        super().__init__(name, *argv, **kwargs)

    def needs_entire_argv(self):
        return True

    def cpp_type(self):
        return f'std::vector<{self.type.cpp_name}*>'


class ArgvObjects(ArgvParameter):
    def cpp_type(self):
        return f'std::vector<{self.type.cpp_name}*>'

    def GenerateCppParamInitialization(self, w, func):
        w.Write(f'{self.cpp_type()} {self.name}(num_args);')
        w.Open('for (int i = 0; i < num_args; ++i) {')
        w.Write(f'{self.name}[i] = reinterpret_cast<'
                f'{self.type.object_struct_name()}*>(args[i])->value;')
        w.Close('}')


class IntegralArgv(ArgvParameter):
    def cpp_type(self):
        return f'std::vector<{self.item_cpp_type()}>'

    def item_cpp_type(self):
        return {
            'i': 'int',
            'u64': 'uint64_t',
            'f32': 'float',
        }[self.type]

    def item_fetch_function(self):
        return {
            'i': 'PyLong_AsLongLong',
        }[self.type]

    def item_fetch_err_value(self):
        return {
            'i': '-1',
        }[self.type]

    def GenerateCppParamInitialization(self, w, func):
        w.Write(f'{self.cpp_type()} {self.name}(num_args);')
        w.Open('for (int i = 0; i < num_args; ++i) {')
        w.Write(f'auto tmp = {self.item_fetch_function()}(args[i]);')
        w.Write(f'if (tmp == {self.item_fetch_err_value()} '
                f'&& PyErr_Occurred() != nullptr) return {func._failure()};')
        w.Write(f'{self.name}[i] = tmp;')
        w.Close('}')



#  This file is part of Leela Chess Zero.
#  Copyright (C) 2020 The LCZero Authors
#
#  Leela Chess is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Leela Chess is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Leela Chess.  If not, see <http://www.gnu.org/licenses/>.
#
#  Additional permission under GNU GPL version 3 section 7
#
#  If you modify this Program, or any covered work, by linking or
#  combining it with NVIDIA Corporation's libraries from the NVIDIA CUDA
#  Toolkit and the NVIDIA CUDA Deep Neural Network library (or a
#  modified version of those libraries), containing parts covered by the
#  terms of the respective license agreement, the licensors of this
#  Program grant you additional permission to convey the resulting work.

from enum import Enum
from .core import PyObject
from .parameters import Parameter
from .retval import RetVal, NoneRetVal
from .exceptions import CppException


class FunctionType(Enum):
    METH_VARARGS = 1
    METH_KEYWORDS = 2
    METH_FASTCALL = 3
    METH_FASTCALL_KEYWORDS = 4
    METH_NOARGS = 5
    METH_O = 6


class Function(PyObject):
    def __init__(self,
                 name,
                 gen_function_name,
                 self_type=None,
                 param_type=None):
        super().__init__(name)
        self.parameters = []
        self.exceptions = []
        self.gen_function_name = gen_function_name
        self.self_type = self_type
        self.param_typ = param_type
        self.retval = NoneRetVal()

    def AddParameter(self, *params):
        for param in params:
            assert isinstance(param, Parameter)
            self.parameters.append(param)
        return self

    def AddRetVal(self, val):
        assert isinstance(self.retval, NoneRetVal)
        assert isinstance(val, RetVal)
        self.retval = val
        return self

    def AddEx(self, ex):
        assert isinstance(ex, CppException)
        self.exceptions.append(ex)
        return self

    def Generate(self, w):
        w.Open(f'{self._return_cpp_type()} '
               f'{self.gen_function_name}({self._generate_params()}) {{')

        self._generate_parse_params(w)
        self.retval.GenerateDeclaration(w)

        if self.exceptions:
            w.Open('try {')

        self._generate_call(w)
        self.retval.GenerateConversion(w)

        if self.exceptions:
            self._handle_exceptions(w)
            w.Close('}')

        w.Write(f'return {self._success()};')
        w.Close('}\n')

    def param_type(self):
        return self.param_typ or Function.TypeFromParameters(self.parameters)

    def function_meth_flags(self):
        return {
            FunctionType.METH_KEYWORDS: 'METH_VARARGS | METH_KEYWORDS',
            FunctionType.METH_NOARGS: 'METH_NOARGS',
            FunctionType.METH_FASTCALL: 'METH_FASTCALL',
        }[self.param_type()]

    def _return_cpp_type(self):
        return 'PyObject*'

    def _generate_parse_params(self, w):
        if self.param_type() == FunctionType.METH_KEYWORDS:
            for param in self.parameters:
                param.GenerateParseTupleSinkDeclaration(w)
            w.Write('const char* keywords[] = {%s};' %
                    ', '.join([f'"{x.name}"'
                               for x in self.parameters] + ['nullptr']))

            def GatherFormatString():
                res = ''
                is_optional = False
                is_keyword_only = False
                for param in self.parameters:
                    if param.optional and not is_optional:
                        is_optional = True
                        res = res + '|'
                    if param.only_keyword and not param.optional:
                        raise ValueError('Non-optional only_keyword.')
                    if param.only_keyword and not is_keyword_only:
                        is_keyword_only = True
                        res = res + '$'
                    if not param.optional and is_optional:
                        raise ValueError('Non-optional after optional.')
                    res += param.parse_tuple_format()
                return res

            params_list = ',\n    '.join([
                'args',
                'kwargs',
                f'"{GatherFormatString()}"',
                'const_cast<char**>(keywords)',
            ] + [
                x for xs in self.parameters
                for x in xs.parse_tuple_sink_list()
            ])
            w.Open(f'if (!PyArg_ParseTupleAndKeywords({params_list})) {{')
            w.Write(f'return {self._failure()};')
            w.Close('}')
            for param in self.parameters:
                param.GenerateCppParamInitialization(w, self)
            return

        if self.param_type() == FunctionType.METH_NOARGS:
            return

        if self.param_type() == FunctionType.METH_FASTCALL:
            assert len(self.parameters) == 1
            param = self.parameters[0]
            assert param.needs_entire_argv()
            param.GenerateCppParamInitialization(w, self)
            return

        raise NotImplementedError('Not implemented function type %d' %
                                  self.param_type())

    def _generate_params(self):
        params = []
        if self.self_type:
            params.append(f'{self.self_type}* self')
        else:
            params.append('void* /* not used */')
        param_type = self.param_type()
        if param_type == FunctionType.METH_VARARGS:
            params.append('PyObject *args')
        elif param_type == FunctionType.METH_KEYWORDS:
            params.append('PyObject *args')
            params.append('PyObject *kwargs')
        elif param_type == FunctionType.METH_FASTCALL:
            params.append('PyObject **args')
            params.append('int num_args')
        elif param_type == FunctionType.METH_FASTCALL_KEYWORDS:
            params.append('PyObject **args')
            params.append('int num_args')
            params.append('PyObject *kwargs')
        elif param_type == FunctionType.METH_NOARGS:
            params.append('PyObject* /* not used */')
        elif param_type == FunctionType.METH_O:
            params.append('PyObject **arg')
        else:
            raise NotImplementedError('Unknown parameter type')
        return ', '.join(params)

    def _list_caller_params(self):
        return ', '.join([x.name_at_caller() for x in self.parameters])

    def _success(self):
        return self.retval.ret_val()

    def _failure(self):
        return 'nullptr'

    def _handle_exceptions(self, w):
        for ex in self.exceptions:
            ex.GenerateHandle(w, self)

    @staticmethod
    def TypeFromParameters(parameters):
        if len(parameters) == 0:
            return FunctionType.METH_NOARGS
        if len(parameters) == 1 and parameters[0].needs_entire_argv():
            return FunctionType.METH_FASTCALL
        return FunctionType.METH_KEYWORDS


class MemberFunction(Function):
    def __init__(self, name, *args, cpp_name=None, **kwargs):
        self.cpp_name = cpp_name or name
        super().__init__(name, *args, **kwargs)

    def _generate_call(self, w):
        if isinstance(self.retval, NoneRetVal):
            w.Write(
                f'self->value->{self.cpp_name}({self._list_caller_params()});')
        else:
            w.Write(f'{self.retval.cpp_type()} '
                    f'{self.retval.cpp_val()} = self->value->'
                    f'{self.cpp_name}({self._list_caller_params()});')


class StaticFunction(Function):
    def __init__(self, name, cpp_type_name, *args, cpp_name=None, **kwargs):
        self.cpp_name = cpp_name or name
        self.cpp_type_name = cpp_type_name
        super().__init__(name, *args, **kwargs)

    def function_meth_flags(self):
        return super().function_meth_flags() + '| METH_STATIC'

    def _generate_call(self, w):
        if isinstance(self.retval, NoneRetVal):
            w.Write(f'{self.cpp_type_name}::{self.cpp_name}'
                    f'({self._list_caller_params()});')
        else:
            w.Write(f'{self.retval.cpp_type()} '
                    f'{self.retval.cpp_val()} = '
                    f'{self.cpp_type_name}::'
                    f'{self.cpp_name}({self._list_caller_params()});')


class Constructor(Function):
    def __init__(self, cpp_type_name, gen_function_name, *args, **kwargs):
        self.cpp_type_name = cpp_type_name
        super().__init__('__init__',
                         gen_function_name=gen_function_name,
                         param_type=FunctionType.METH_KEYWORDS,
                         *args,
                         **kwargs)

    def _generate_call(self, w):
        w.Write(f'self->value = new '
                f'{self.cpp_type_name}({self._list_caller_params()});')

    def _return_cpp_type(self):
        return 'int'

    def _success(self):
        return '0'

    def _failure(self):
        return '-1'


class DisabledConstructor(Function):
    def __init__(self, gen_function_name):
        super().__init__('__init__',
                         gen_function_name=gen_function_name,
                         param_type=FunctionType.METH_KEYWORDS)

    def Generate(self, w):
        w.Open(f'{self._return_cpp_type()} '
               f'{self.gen_function_name}({self._generate_params()}) {{')
        w.Write('PyErr_SetString(PyExc_TypeError, '
                '"Not possible to create instances of this type.");')
        w.Write(f'return {self._failure()};')
        w.Close('}\n')

    def _return_cpp_type(self):
        return 'int'

    def _success(self):
        return '0'

    def _failure(self):
        return '-1'